@@ -30,6 +30,7 @@
 #include "caml/misc.h"
 #include "caml/mlvalues.h"
 #include "caml/signals.h"
+#include "caml/eventlog.h"
 #ifdef NATIVE_CODE
 #include "caml/stack.h"
 #else
@@ -154,7 +155,8 @@ static value heap_stats (int returnstats)
           ++ fragments;
           CAMLassert (prev_hp == NULL
                       || Color_hp (prev_hp) != Caml_blue
-                      || cur_hp == (header_t *) caml_gc_sweep_hp);
+                      || cur_hp == (header_t *) caml_gc_sweep_hp
+                      || Wosize_hp (prev_hp) == Max_wosize);
         }else{
           if (caml_gc_phase == Phase_sweep
               && cur_hp >= (header_t *) caml_gc_sweep_hp){
@@ -172,7 +174,7 @@ static value heap_stats (int returnstats)
           }
         }
         break;
-      case Caml_gray: case Caml_black:
+      case Caml_black:
         CAMLassert (Wosize_hd (cur_hd) > 0);
         ++ live_blocks;
         live_words += Whsize_hd (cur_hd);
@@ -232,9 +234,10 @@ static value heap_stats (int returnstats)
     intnat majcoll = Caml_state->stat_major_collections;
     intnat heap_words = Caml_state->stat_heap_wsz;
     intnat cpct = Caml_state->stat_compactions;
+    intnat forcmajcoll = Caml_state->stat_forced_major_collections;
     intnat top_heap_words = Caml_state->stat_top_heap_wsz;
 
-    res = caml_alloc_tuple (16);
+    res = caml_alloc_tuple (17);
     Store_field (res, 0, caml_copy_double (minwords));
     Store_field (res, 1, caml_copy_double (prowords));
     Store_field (res, 2, caml_copy_double (majwords));
@@ -251,6 +254,7 @@ static value heap_stats (int returnstats)
     Store_field (res, 13, Val_long (cpct));
     Store_field (res, 14, Val_long (top_heap_words));
     Store_field (res, 15, Val_long (caml_stack_usage()));
+    Store_field (res, 16, Val_long (forcmajcoll));
     CAMLreturn (res);
   }else{
     CAMLreturn (Val_unit);
@@ -267,10 +271,10 @@ void caml_heap_check (void)
 CAMLprim value caml_gc_stat(value v)
 {
   value result;
-  CAML_INSTR_SETUP (tmr, "");
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_STAT);
   CAMLassert (v == Val_unit);
   result = heap_stats (1);
-  CAML_INSTR_TIME (tmr, "explicit/gc_stat");
+  CAML_EV_END(EV_EXPLICIT_GC_STAT);
   return result;
 }
 
@@ -291,9 +295,10 @@ CAMLprim value caml_gc_quick_stat(value v)
   intnat heap_words = Caml_state->stat_heap_wsz;
   intnat top_heap_words = Caml_state->stat_top_heap_wsz;
   intnat cpct = Caml_state->stat_compactions;
+  intnat forcmajcoll = Caml_state->stat_forced_major_collections;
   intnat heap_chunks = Caml_state->stat_heap_chunks;
 
-  res = caml_alloc_tuple (16);
+  res = caml_alloc_tuple (17);
   Store_field (res, 0, caml_copy_double (minwords));
   Store_field (res, 1, caml_copy_double (prowords));
   Store_field (res, 2, caml_copy_double (majwords));
@@ -310,6 +315,7 @@ CAMLprim value caml_gc_quick_stat(value v)
   Store_field (res, 13, Val_long (cpct));
   Store_field (res, 14, Val_long (top_heap_words));
   Store_field (res, 15, Val_long (caml_stack_usage()));
+  Store_field (res, 16, Val_long (forcmajcoll));
   CAMLreturn (res);
 }
 
@@ -422,7 +428,7 @@ CAMLprim value caml_gc_set(value v)
   asize_t newminwsz;
   uintnat newpolicy;
   uintnat new_custom_maj, new_custom_min, new_custom_sz;
-  CAML_INSTR_SETUP (tmr, "");
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_SET);
 
   caml_verb_gc = Long_val (Field (v, 3));
 
@@ -470,21 +476,21 @@ CAMLprim value caml_gc_set(value v)
 
   /* These fields were added in 4.08.0. */
   if (Wosize_val (v) >= 11){
-    new_custom_maj = norm_custom_maj (Field (v, 8));
+    new_custom_maj = norm_custom_maj (Long_val (Field (v, 8)));
     if (new_custom_maj != caml_custom_major_ratio){
       caml_custom_major_ratio = new_custom_maj;
       caml_gc_message (0x20, "New custom major ratio: %"
                        ARCH_INTNAT_PRINTF_FORMAT "u%%\n",
                        caml_custom_major_ratio);
     }
-    new_custom_min = norm_custom_min (Field (v, 9));
+    new_custom_min = norm_custom_min (Long_val (Field (v, 9)));
     if (new_custom_min != caml_custom_minor_ratio){
       caml_custom_minor_ratio = new_custom_min;
       caml_gc_message (0x20, "New custom minor ratio: %"
                        ARCH_INTNAT_PRINTF_FORMAT "u%%\n",
                        caml_custom_minor_ratio);
     }
-    new_custom_sz = Field (v, 10);
+    new_custom_sz = Long_val (Field (v, 10));
     if (new_custom_sz != caml_custom_minor_max_bsz){
       caml_custom_minor_max_bsz = new_custom_sz;
       caml_gc_message (0x20, "New custom minor size limit: %"
@@ -501,8 +507,10 @@ CAMLprim value caml_gc_set(value v)
   newpolicy = Long_val (Field (v, 6));
   if (newpolicy != caml_allocation_policy){
     caml_empty_minor_heap ();
+    caml_gc_message (0x1, "Full major GC cycle (changing allocation policy)\n");
     caml_finish_major_cycle ();
     caml_finish_major_cycle ();
+    ++ Caml_state->stat_forced_major_collections;
     caml_compact_heap (newpolicy);
     caml_gc_message (0x20, "New allocation policy: %"
                      ARCH_INTNAT_PRINTF_FORMAT "u\n", newpolicy);
@@ -514,7 +522,7 @@ CAMLprim value caml_gc_set(value v)
                      ARCH_SIZET_PRINTF_FORMAT "uk words\n", newminwsz / 1024);
     caml_set_minor_heap_size (Bsize_wsize (newminwsz));
   }
-  CAML_INSTR_TIME (tmr, "explicit/gc_set");
+  CAML_EV_END(EV_EXPLICIT_GC_SET);
 
   /* The compaction may have triggered some finalizers that we need to call. */
   caml_process_pending_actions();
@@ -524,12 +532,15 @@ CAMLprim value caml_gc_set(value v)
 
 CAMLprim value caml_gc_minor(value v)
 {
-  CAML_INSTR_SETUP (tmr, "");
+  value exn;
+
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_MINOR);
   CAMLassert (v == Val_unit);
   caml_request_minor_gc ();
   // call the gc and call finalisers
-  caml_process_pending_actions();
-  CAML_INSTR_TIME (tmr, "explicit/gc_minor");
+  exn = caml_process_pending_actions_exn();
+  CAML_EV_END(EV_EXPLICIT_GC_MINOR);
+  caml_raise_if_exception(exn);
   return Val_unit;
 }
 
@@ -550,60 +561,90 @@ static void test_and_compact (void)
 
 CAMLprim value caml_gc_major(value v)
 {
-  CAML_INSTR_SETUP (tmr, "");
+  value exn;
+
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_MAJOR);
   CAMLassert (v == Val_unit);
-  caml_gc_message (0x1, "Major GC cycle requested\n");
+  caml_gc_message (0x1, "Finishing major GC cycle (requested by user)\n");
   caml_empty_minor_heap ();
   caml_finish_major_cycle ();
   test_and_compact ();
   // call finalisers
-  caml_process_pending_actions();
-  CAML_INSTR_TIME (tmr, "explicit/gc_major");
+  exn = caml_process_pending_actions_exn();
+  CAML_EV_END(EV_EXPLICIT_GC_MAJOR);
+  caml_raise_if_exception(exn);
   return Val_unit;
 }
 
 CAMLprim value caml_gc_full_major(value v)
 {
-  CAML_INSTR_SETUP (tmr, "");
+  value exn;
+
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_FULL_MAJOR);
   CAMLassert (v == Val_unit);
-  caml_gc_message (0x1, "Full major GC cycle requested\n");
+  caml_gc_message (0x1, "Full major GC cycle (requested by user)\n");
   caml_empty_minor_heap ();
   caml_finish_major_cycle ();
   // call finalisers
-  caml_process_pending_actions();
+  exn = caml_process_pending_actions_exn();
+  if (Is_exception_result(exn)) goto cleanup;
   caml_empty_minor_heap ();
   caml_finish_major_cycle ();
+  ++ Caml_state->stat_forced_major_collections;
   test_and_compact ();
   // call finalisers
-  caml_process_pending_actions();
-  CAML_INSTR_TIME (tmr, "explicit/gc_full_major");
+  exn = caml_process_pending_actions_exn();
+
+cleanup:
+  CAML_EV_END(EV_EXPLICIT_GC_FULL_MAJOR);
+  caml_raise_if_exception(exn);
+
   return Val_unit;
 }
 
 CAMLprim value caml_gc_major_slice (value v)
 {
-  CAML_INSTR_SETUP (tmr, "");
+  value exn = Val_unit;
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_MAJOR_SLICE);
   CAMLassert (Is_long (v));
-  caml_major_collection_slice (Long_val (v));
-  CAML_INSTR_TIME (tmr, "explicit/gc_major_slice");
+  if (caml_gc_phase == Phase_idle){
+    /* We need to start a new major GC cycle. Go through the pending_action
+       machinery. */
+    caml_request_major_slice ();
+    exn = caml_process_pending_actions_exn ();
+      /* Calls the major GC without passing [v] but the initial slice
+         ignores this parameter anyway. */
+  }else{
+    caml_major_collection_slice (Long_val (v));
+  }
+  CAML_EV_END(EV_EXPLICIT_GC_MAJOR_SLICE);
+  caml_raise_if_exception (exn);
   return Val_long (0);
 }
 
 CAMLprim value caml_gc_compaction(value v)
 {
-  CAML_INSTR_SETUP (tmr, "");
+  value exn;
+
+  CAML_EV_BEGIN(EV_EXPLICIT_GC_COMPACT);
   CAMLassert (v == Val_unit);
   caml_gc_message (0x10, "Heap compaction requested\n");
   caml_empty_minor_heap ();
+  caml_gc_message (0x1, "Full major GC cycle (compaction)\n");
   caml_finish_major_cycle ();
   // call finalisers
-  caml_process_pending_actions();
+  exn = caml_process_pending_actions_exn();
+  if (Is_exception_result(exn)) goto cleanup;
   caml_empty_minor_heap ();
   caml_finish_major_cycle ();
+  ++ Caml_state->stat_forced_major_collections;
   caml_compact_heap (-1);
   // call finalisers
-  caml_process_pending_actions();
-  CAML_INSTR_TIME (tmr, "explicit/gc_compact");
+  exn = caml_process_pending_actions_exn();
+
+ cleanup:
+  CAML_EV_END(EV_EXPLICIT_GC_COMPACT);
+  caml_raise_if_exception(exn);
   return Val_unit;
 }
 
@@ -638,14 +679,13 @@ void caml_init_gc (uintnat minor_size, uintnat major_size,
                    uintnat major_incr, uintnat percent_fr,
                    uintnat percent_m, uintnat window,
                    uintnat custom_maj, uintnat custom_min,
-                   uintnat custom_bsz)
+                   uintnat custom_bsz, uintnat policy)
 {
   uintnat major_bsize;
   if (major_size < Heap_chunk_min) major_size = Heap_chunk_min;
   major_bsize = Bsize_wsize(major_size);
   major_bsize = ((major_bsize + Page_size - 1) >> Page_log) << Page_log;
 
-  caml_instr_init ();
   if (caml_page_table_initialize(Bsize_wsize(minor_size) + major_bsize)){
     caml_fatal_error ("cannot initialize page table");
   }
@@ -653,6 +693,7 @@ void caml_init_gc (uintnat minor_size, uintnat major_size,
   caml_major_heap_increment = major_incr;
   caml_percent_free = norm_pfree (percent_fr);
   caml_percent_max = norm_pmax (percent_m);
+  caml_set_allocation_policy (policy);
   caml_init_major_heap (major_bsize);
   caml_major_window = norm_window (window);
   caml_custom_major_ratio = norm_custom_maj (custom_maj);
