@@ -19,46 +19,53 @@ open Asttypes
 open Path
 open Types
 
-
-let rec scrape env mty =
+let rec scrape_lazy env mty =
+  let open Subst.Lazy in
   match mty with
-    Mty_ident p ->
+    MtyL_ident p ->
       begin try
-        scrape env (Env.find_modtype_expansion p env)
+        scrape_lazy env (Env.find_modtype_expansion_lazy p env)
       with Not_found ->
         mty
       end
   | _ -> mty
 
+let scrape env mty =
+  match mty with
+    Mty_ident p ->
+     Subst.Lazy.force_modtype (scrape_lazy env (MtyL_ident p))
+  | _ -> mty
+
 let freshen ~scope mty =
   Subst.modtype (Rescope scope) Subst.identity mty
 
-let rec strengthen ~aliasable env mty p =
-  match scrape env mty with
-    Mty_signature sg ->
-      Mty_signature(strengthen_sig ~aliasable env sg p)
-  | Mty_functor(Named (Some param, arg), res)
+let rec strengthen_lazy ~aliasable env mty p =
+  let open Subst.Lazy in
+  match scrape_lazy env mty with
+    MtyL_signature sg ->
+      MtyL_signature(strengthen_lazy_sig ~aliasable env sg p)
+  | MtyL_functor(Named (Some param, arg), res)
     when !Clflags.applicative_functors ->
-      Mty_functor(Named (Some param, arg),
-        strengthen ~aliasable:false env res (Papply(p, Pident param)))
-  | Mty_functor(Named (None, arg), res)
+      MtyL_functor(Named (Some param, arg),
+        strengthen_lazy ~aliasable:false env res (Papply(p, Pident param)))
+  | MtyL_functor(Named (None, arg), res)
     when !Clflags.applicative_functors ->
       let param = Ident.create_scoped ~scope:(Path.scope p) "Arg" in
-      Mty_functor(Named (Some param, arg),
-        strengthen ~aliasable:false env res (Papply(p, Pident param)))
+      MtyL_functor(Named (Some param, arg),
+        strengthen_lazy ~aliasable:false env res (Papply(p, Pident param)))
   | mty ->
       mty
 
-and strengthen_sig ~aliasable env sg p =
+and strengthen_lazy_sig' ~aliasable env sg p =
+  let open Subst.Lazy in
   match sg with
     [] -> []
-  | (Sig_value(_, _, _) as sigelt) :: rem ->
-      sigelt :: strengthen_sig ~aliasable env rem p
-  | Sig_type(id, {type_kind=Type_abstract}, _, _) ::
-    (Sig_type(id', {type_private=Private}, _, _) :: _ as rem)
-    when Ident.name id = Ident.name id' ^ "#row" ->
-      strengthen_sig ~aliasable env rem p
-  | Sig_type(id, decl, rs, vis) :: rem ->
+  | (SigL_value(_, _, _) as sigelt) :: rem ->
+      sigelt :: strengthen_lazy_sig' ~aliasable env rem p
+  | SigL_type(id, {type_kind=Type_abstract}, _, _) :: rem
+    when Btype.is_row_name (Ident.name id) ->
+      strengthen_lazy_sig' ~aliasable env rem p
+  | SigL_type(id, decl, rs, vis) :: rem ->
       let newdecl =
         match decl.type_manifest, decl.type_private, decl.type_kind with
           Some _, Public, _ -> decl
@@ -72,40 +79,60 @@ and strengthen_sig ~aliasable env sg p =
             else
               { decl with type_manifest = manif }
       in
-      Sig_type(id, newdecl, rs, vis) :: strengthen_sig ~aliasable env rem p
-  | (Sig_typext _ as sigelt) :: rem ->
-      sigelt :: strengthen_sig ~aliasable env rem p
-  | Sig_module(id, pres, md, rs, vis) :: rem ->
+      SigL_type(id, newdecl, rs, vis) ::
+        strengthen_lazy_sig' ~aliasable env rem p
+  | (SigL_typext _ as sigelt) :: rem ->
+      sigelt :: strengthen_lazy_sig' ~aliasable env rem p
+  | SigL_module(id, pres, md, rs, vis) :: rem ->
       let str =
-        strengthen_decl ~aliasable env md (Pdot(p, Ident.name id))
+        strengthen_lazy_decl ~aliasable env md (Pdot(p, Ident.name id))
       in
-      Sig_module(id, pres, str, rs, vis)
-      :: strengthen_sig ~aliasable
-        (Env.add_module_declaration ~check:false id pres md env) rem p
+      let env =
+        Env.add_module_declaration_lazy ~update_summary:false id pres md env in
+      SigL_module(id, pres, str, rs, vis)
+      :: strengthen_lazy_sig' ~aliasable env rem p
       (* Need to add the module in case it defines manifest module types *)
-  | Sig_modtype(id, decl, vis) :: rem ->
+  | SigL_modtype(id, decl, vis) :: rem ->
       let newdecl =
-        match decl.mtd_type with
-          None ->
-            {decl with mtd_type = Some(Mty_ident(Pdot(p,Ident.name id)))}
-        | Some _ ->
+        match decl.mtdl_type with
+        | Some _ when not aliasable ->
+            (* [not alisable] condition needed because of recursive modules.
+               See [Typemod.check_recmodule_inclusion]. *)
             decl
+        | _ ->
+            {decl with mtdl_type = Some(MtyL_ident(Pdot(p,Ident.name id)))}
       in
-      Sig_modtype(id, newdecl, vis) ::
-      strengthen_sig ~aliasable (Env.add_modtype id decl env) rem p
+      let env = Env.add_modtype_lazy ~update_summary:false id decl env in
+      SigL_modtype(id, newdecl, vis) ::
+      strengthen_lazy_sig' ~aliasable env rem p
       (* Need to add the module type in case it is manifest *)
-  | (Sig_class _ as sigelt) :: rem ->
-      sigelt :: strengthen_sig ~aliasable env rem p
-  | (Sig_class_type _ as sigelt) :: rem ->
-      sigelt :: strengthen_sig ~aliasable env rem p
+  | (SigL_class _ as sigelt) :: rem ->
+      sigelt :: strengthen_lazy_sig' ~aliasable env rem p
+  | (SigL_class_type _ as sigelt) :: rem ->
+      sigelt :: strengthen_lazy_sig' ~aliasable env rem p
 
-and strengthen_decl ~aliasable env md p =
-  match md.md_type with
-  | Mty_alias _ -> md
-  | _ when aliasable -> {md with md_type = Mty_alias p}
-  | mty -> {md with md_type = strengthen ~aliasable env mty p}
+and strengthen_lazy_sig ~aliasable env sg p =
+  let sg = Subst.Lazy.force_signature_once sg in
+  let sg = strengthen_lazy_sig' ~aliasable env sg p in
+  Subst.Lazy.of_signature_items sg
 
-let () = Env.strengthen := strengthen
+and strengthen_lazy_decl ~aliasable env md p =
+  let open Subst.Lazy in
+  match md.mdl_type with
+  | MtyL_alias _ -> md
+  | _ when aliasable -> {md with mdl_type = MtyL_alias p}
+  | mty -> {md with mdl_type = strengthen_lazy ~aliasable env mty p}
+
+let () = Env.strengthen := strengthen_lazy
+
+let strengthen ~aliasable env mty p =
+  let mty = strengthen_lazy ~aliasable env (Subst.Lazy.of_modtype mty) p in
+  Subst.Lazy.force_modtype mty
+
+let strengthen_decl ~aliasable env md p =
+  let md = strengthen_lazy_decl ~aliasable env
+             (Subst.Lazy.of_module_decl md) p in
+  Subst.Lazy.force_module_decl md
 
 let rec make_aliases_absent pres mty =
   match mty with
@@ -213,7 +240,7 @@ and nondep_sig_item env va ids = function
       with Ctype.Nondep_cannot_erase _ as exn ->
         match va with
           Co -> Sig_modtype(id, {mtd_type=None; mtd_loc=Location.none;
-                                 mtd_attributes=[]}, vis)
+                                 mtd_attributes=[]; mtd_uid = d.mtd_uid}, vis)
         | _  -> raise exn
       end
   | Sig_class(id, d, rs, vis) ->
@@ -222,6 +249,8 @@ and nondep_sig_item env va ids = function
       Sig_class_type(id, Ctype.nondep_cltype_declaration env ids d, rs, vis)
 
 and nondep_sig env va ids sg =
+  let scope = Ctype.create_scope () in
+  let sg, env = Env.enter_signature ~scope sg env in
   List.map (nondep_sig_item env va ids) sg
 
 and nondep_modtype_decl env ids mtd =
@@ -234,33 +263,35 @@ let enrich_typedecl env p id decl =
   match decl.type_manifest with
     Some _ -> decl
   | None ->
-      try
-        let orig_decl = Env.find_type p env in
+    match Env.find_type p env with
+    | exception Not_found -> decl
+        (* Type which was not present in the signature, so we don't have
+           anything to do. *)
+    | orig_decl ->
         if decl.type_arity <> orig_decl.type_arity then
           decl
-        else
+        else begin
           let orig_ty =
-            Ctype.reify_univars
+            Ctype.reify_univars env
               (Btype.newgenty(Tconstr(p, orig_decl.type_params, ref Mnil)))
           in
           let new_ty =
-            Ctype.reify_univars
+            Ctype.reify_univars env
               (Btype.newgenty(Tconstr(Pident id, decl.type_params, ref Mnil)))
           in
           let env = Env.add_type ~check:false id decl env in
-          Ctype.mcomp env orig_ty new_ty;
-          let orig_ty =
-            Btype.newgenty(Tconstr(p, decl.type_params, ref Mnil))
-          in
-          {decl with type_manifest = Some orig_ty}
-      with Not_found | Ctype.Unify _ ->
-        (* - Not_found: type which was not present in the signature, so we don't
-           have anything to do.
-           - Unify: the current declaration is not compatible with the one we
-           got from the signature. We should just fail now, but then, we could
-           also have failed if the arities of the two decls were different,
-           which we didn't. *)
-        decl
+          match Ctype.mcomp env orig_ty new_ty with
+          | exception Ctype.Incompatible -> decl
+              (* The current declaration is not compatible with the one we got
+                 from the signature. We should just fail now, but then, we could
+                 also have failed if the arities of the two decls were
+                 different, which we didn't. *)
+          | () ->
+              let orig_ty =
+                Btype.newgenty(Tconstr(p, decl.type_params, ref Mnil))
+              in
+              {decl with type_manifest = Some orig_ty}
+        end
 
 let rec enrich_modtype env p mty =
   match mty with
@@ -515,9 +546,9 @@ let scrape_for_type_of ~remove_aliases env mty =
 let lower_nongen nglev mty =
   let open Btype in
   let it_type_expr it ty =
-    let ty = repr ty in
-    match ty with
-      {desc=Tvar _; level} ->
+    match get_desc ty with
+      Tvar _ ->
+        let level = get_level ty in
         if level < generic_level && level > nglev then set_level ty nglev
     | _ ->
         type_iterators.it_type_expr it ty
